#include <math.h>
#include <float.h>
/**
# Un programa para calcular los tres eigenvalores 
# menores de una matrix de tamano n x n, y el correspondiente
# eigenvector (o vector singular) tal que min(li) y li > 1e-8, para i=1,2,3,
# donde li es cada uno de los eigenvalores calculados
#
# Ideas tomadas de:
# [1] G.H. Golun and C.F. Van Loan, Matrix Computations, 3rd edition
#   chapter 8.5, page 439 
# [2] W. Barth, R.S. Martin and J.H. Wilkinson, Calculation of the
#   eigenvalues of a symmetric tridiagonal matrix by the method of
#   bisection, Numerische Mathematik 9, pp. 386-393, 1967
#
#  Fraga 24/01/2014
#
# 1. Se calcula B = A^T * A
# 2. Se reduce B a la forma tridiagonal
# 3. Se hallan los extremos de los valores de
#    todos los eigenvalores y se aplica [2]
##### Ok al 25/01/2014
# 4. Se hallan los tres eigenvalores ms pequeos
# 5. Se busca el eigenvalor ms pequeos y mayor a 1e-8, este es l
# 6. Se calcula el eigenvector correspondiente
#    al eigenvalor ms pequeo por el mtodo
#    de iteracin inversa, usando la matriz B:
#     (B-lI)x = b0
# 
#    b = [1, 1, ..., 1]^T; b0 = b/norm(b);
#    Se resuelve usando la descomposicin QR de (B-lI)   
#
##### Ok al 28/01/2014 en Perl
   Versin en C 15/05/2014
**/

/**
 Realiza la transformacin Householder
 A es de tamao nxn, 
 v es un vector de tamao n 
**/
double houseHolder( double **A,  int n, int k, double *v )
{
	double sigma, tmp;
	double beta, mu, val;
	int i, k1;

	// Cambiado: house( A( k+1:n-1, k )
    // v(0) = A( k+1, k )
    // v(1) = A( k+2, k )
	// v(3) = A( k+3, k ) 
	//
    // v(m) = A( n-1, k )
	//
	// Si v(0,...,n-1)
	//#Si v(k+1,...,n-1)

	sigma = 0.0;
	for( i=k+2; i<n; i++ ) {
		tmp = A[i][k]; 
		sigma += tmp*tmp;
	}
	
	v[k+1] = 1.0;
	for( i=k+2; i<n; i++ ) {
		v[i] = A[i][k];
	}
	if ( sigma < 1e-10 ) {
		beta = 0.0;
	}
	else {
		k1 = k+1;
		tmp = A[k1][k];
		mu = sqrt( tmp * tmp + sigma );
		if ( tmp <= 0.0 ) {
			v[k1] = A[k1][k] - mu;
		}
		else {
			v[k1] = -sigma/(A[k1][k] + mu);
		}
		tmp = v[k1];
		val = tmp * tmp;
		beta = 2.0*val/(sigma+val); 
		val = v[k1];
		for( i=k1; i<n; i++ ) {
			v[i] /= val;
		}
	}
	return( beta );
}

/**
   Funcin que reduce la matriz A, de tamao n x n,
 a una matriz tridiagonal. La matriz de salida se sobreescribe en A
**/
void tridiagonal( double **A, int n, double *vec )
{
	double *v, *p, *w;
	double beta, suma, val;
	int i, j, k, k1;

	v=vec;
	p = &vec[n];
	w = &vec[2*n];

	for( k=0; k<n-2; k++ ) {
		//                                     n = 5,  m =  n - (k+1) +1 = n-k
		// Original: house( A( k+1:n , k )  si k=1, m = 5 - 1 = 4
		// Cambiado: house( A( k+1:n-1, k ) si k=0, m = n-1-(k+1)+1 = n-k-1 = 4
		beta = houseHolder( A, n, k, v );

		k1 = k + 1;
		for( i=k1; i<n; i++ ) { 
			suma = 0.0;
			for( j=k+1; j<n; j++ ) { 
				suma += A[i][j] * v[j];
			}
			p[i] = beta*suma;
		}
		val = 0.0;
		for( j=k1; j<n; j++ ) 
			val += p[j] * v[j];

		val *= beta / 2.0;

		for( j=k1; j<n; j++ ) 
			w[j] = p[j] - val * v[j]; 

		suma = 0.0;
		for( j=k1; j<n; j++ ) { 
			val = A[j][k];
			suma += val * val;
		}
		val = sqrt( suma );

		// si k=2
		A[k1][k] = val;
		A[k][k1] = val;

		for( i=k+1; i<n; i++ ) { 
			for( j=k+1; j<n; j++ ) { 
				A[i][j] -=  v[i]*w[j] + w[i]*v[j];
			}
		}
	}
}

/**
# Rutina que multiplica A^T * A
# Solo es necesario calcular la porcin superior (o inferior)
# de la matriz dado que el resultado es una matriz simtrica
# Args: S = E^T * E, y n, el tamao de las matrices es n x n
**/
void multiplicaATA( double **S, double **E, int n )
{
	int i, j, k;
	double suma;

	for( i=0; i<n; i++ ) {
		suma = 0.0;
		for( k=0; k<n; k++ ) {
			suma += E[k][i] * E[k][i];
		}
		S[i][i] = suma;
		for( j=i+1; j<n; j++ ) {
			suma = 0.0;
			for( k=0; k<n; k++ ) {
				suma += E[k][i] * E[k][j];
			}
			S[i][j] = suma;
			S[j][i] = suma;
		}
	}
}

/**
# Rutina que encuentra los eigenvalores de una matriz tridiagonal
# Args: M, la matriz tridiagonal de entrada
#       n, el tamao de la matriz de n x n
#      eigen, referencia para almacenar los eigenvalores
#      m1, m2, los eigenvalores deseados, l_m1 < l_m1+1 < l_m1+2 < ... < l_m2
**/
void biseccion( double **M, int n, double *eigen, int m1, int m2, double *vec )
{
	int i, k, z, a;
	double *beta, tmp;
	double xmin, xmax;
	double h, v;
	double x0, *wu;
	double xu, x1, q;
	double eps1, eps2;

	beta = vec;
	beta[0] = 0.0;
	for( i=n-1; i>=1; i-- ){
		tmp = M[i][i-1];
		beta[i] = tmp*tmp;
	}
	// Calculo of xmin y xmax
	k = n-1;
	xmin = M[k][k] - fabs( M[k][k-1] );
	xmax = M[k][k] + fabs( M[k][k-1] );

	for( i=n-2; i>0; i-- ){
		h = fabs(M[i][i-1]) + fabs(M[i+1][i]);
		v = M[i][i] + h;
		if ( v > xmax ) {
			xmax = v;
		}
		v = M[i][i] - h;
		if ( v < xmin ) {
			xmin = v;
		}
	}
	// my $DBL_EPSILON = 2.22045E-16;
	eps1=1e-8;
	if( xmin + xmax > 0.0 ) {
		eps2 = DBL_EPSILON*xmax;
	}
	else {
		eps2 = DBL_EPSILON*(-xmin);
	}
	if( eps1 <= 0 ) {
		eps1 = eps2;
	}
	eps2 = 0.5*eps1 + 7.0*eps2;

	x0 = xmax;
	wu = &vec[n];
	for( i=m1; i<=m2; i++ ) {
		eigen[i] = xmax;
		wu[i] = xmin;
	}
	z = 0;

	//  Ciclo para calcular el k-simo eigenvalor
	for( k=m2; k>=m1; k-- ) {

		xu = xmin;
		for( i=k; i>=m1; i-- ) {
			if( xu < wu[i] ){
				xu = wu[i]; 
				break;
			}
		}
		if( x0 > eigen[k] ) {
			x0 = eigen[k];
		}
	
		while(  x0-xu > 2.0*DBL_EPSILON*(fabs(xu)+fabs(x0))+eps1 ) {
			z++;

			x1 = (xu+x0)/2.0;
			a = 0;
			q = 1.0;
			for( i=0; i<n; i++ ) {
				if ( fabs(q) > 1e-10 ) {
					v = beta[i]/q;
				}
				else {
					if ( i==0 ){ 
						v = 0.0;
					}
					else{
						v = fabs( M[i-1][i]/DBL_EPSILON );
					}
				}
				q = M[i][i] - x1 - v; 
				if( q<0.0 ) {
					a++;
				}
			}
			if( a<k ) {
				if( a<m1 ) {
					xu = wu[m1] = x1;
				}
				else{
					xu = wu[a+1] = x1;
					if( eigen[a] > x1 ){
						eigen[a] = x1;
					}
				}
			}
			else {
				x0 = x1;
			}
		}
		eigen[k] = (x0+xu)/2.0; 
	}
	eigen[0] = z;
}

/**
  Subrutina que calcula descomposicin QR
  A = QR, A es de tamao m x n
  No retorna el valor de  Q
**/
void descomposicionQR( double **A, int m, int n, double **R, double *vec )
{
	int i, j, k;
	double sum, tmp;

	for( k=0; k<n; k++ ) {

		sum = 0.0;
		for( i=0; i<m; i++ ) {
			tmp = A[i][k];
			sum += tmp * tmp;
		}
		sum = sqrt( sum );
		R[k][k] = sum;
		for( i=0; i<m; i++ )
			vec[i] = A[i][k] / sum;

		for( j=k+1; j<n; j++ ) {
			sum = 0.0;
			for( i=0; i<m; i++ ) 
				sum += vec[i] * A[i][j];

			R[k][j] = sum;

			for( i=0; i<m; i++ ) 
				A[i][j] -= vec[i] * sum;
		}
	}
}

/**
 Si se tiene el sistema Ax = b, A es de tamao n x n, 
se realiza la matriz aumentada [A|b] y se realiza
la descomposicin QR, se obtiene Q[ R | Q^T b ]
Esta rutina obtiene x a partir de la matriz Rp = [ R | Q^T b ]
**/
void resuelve( double **Rp, int n, double *x )
{
	int i, j;
	double suma;

	j = n-1;

	x[j] = Rp[j][n] / Rp[j][j];

	for( i=n-2; i>=0; i-- ) {
		suma = 0.0;
		for( j=i+1; j<n; j++ )
			suma += Rp[i][j] * x[j];

		x[i] = (Rp[i][n] - suma)/Rp[i][i];
	}
}

